DEFAULT_SELECTIVE_DOMAIN = False
SELECTIVE_DOMAIN = "enable_selective_domain"

SSL_CLIENT_CERT_FIELD = "ssl_client_cert"
SSL_CLIENT_KEY_FIELD = "ssl_client_key"

# Special marker key used by EndpointConfig to indicate 449 status
# without raising an exception
MISSING_DOMAIN_MARKER = "missing_domain"
