CONFIG_API_KEY_KEY = "api_key"
CONFIG_TOKEN_KEY = "token"
CONFIG_OAUTH_KEY = "oauth"
CONFIG_HEADER_NAME_KEY = "header_name"
CONFIG_HEADER_FORMAT_KEY = "header_format"
CONFIG_CLIENT_ID_KEY = "client_id"
CONFIG_CLIENT_SECRET_KEY = "client_secret"
CONFIG_TOKEN_URL_KEY = "token_url"
CONFIG_SCOPE_KEY = "scope"
CONFIG_AUDIENCE_KEY = "audience"
CONFIG_GRANT_TYPE_KEY = "grant_type"
CONFIG_TIMEOUT_KEY = "timeout"
CONFIG_MODULE_KEY = "module"
