import rasa.e2e_test.constants as constants

KEY_FIXTURES = constants.KEY_FIXTURES
KEY_METADATA = constants.KEY_METADATA
KEY_STUB_CUSTOM_ACTIONS = constants.KEY_STUB_CUSTOM_ACTIONS
KEY_USER_INPUT = constants.KEY_USER_INPUT
KEY_BOT_INPUT = constants.KEY_BOT_INPUT
KEY_BOT_UTTERED = constants.KEY_BOT_UTTERED
KEY_STEPS = constants.KEY_STEPS
KEY_TEST_CASE = constants.KEY_TEST_CASE
KEY_TEST_CASES = constants.KEY_TEST_CASES
KEY_COMMANDS = "commands"
KEY_ADDITIONAL_COMMANDS = "additional_commands"
KEY_REMOVE_DEFAULT_COMMANDS = "remove_default_commands"

ACTOR_USER = "user"
ACTOR_BOT = "bot"

DEFAULT_INPUT_TESTS_PATH = "dialogue_understanding_tests/"
SCHEMA_FILE_PATH = "dialogue_understanding_test/du_test_schema.yml"

PLACEHOLDER_GENERATED_ANSWER_TEMPLATE = "placeholder_generated_answer"
